﻿using System;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using TSF.InteropTypes;
using TSF.SearchCandidateProviderInternal;
using TSF.TypeLib;
using WindowsInput;

namespace SoftKeyboardDemo
{
    public partial class Keyboard : UserControl, ITfUIElementSink, ITfTextEditSink, ITfThreadMgrEventSink
    {
        #region Click Command

        private static readonly RoutedCommand _clickCmd = new RoutedCommand("clickCmd", typeof(Keyboard));

        public static RoutedCommand ClickCmd
        {
            get { return _clickCmd; }
        }

        #endregion

        private ComReleaser _releaser = new ComReleaser();
        private ITfThreadMgrEx _threadMgr;
        private ITfUIElementMgr _uiElementMgr;

        private uint _textEditSinkCookie = TextFrameworkDeclarations.TF_INVALID_COOKIE;
        private uint _threadMgrEventSinkCookie;

        private ITfContext _textEditSinkContext;

        public ObservableCollection<string> Candidates
        {
            get { return (ObservableCollection<string>)GetValue(CandidatesProperty); }
            set { SetValue(CandidatesProperty, value); }
        }

        public static readonly DependencyProperty CandidatesProperty =
            DependencyProperty.Register("Candidates", typeof(ObservableCollection<string>), typeof(Keyboard));

        public string CompositionString
        {
            get { return (string)GetValue(CompositionStringProperty); }
            set { SetValue(CompositionStringProperty, value); }
        }

        public static readonly DependencyProperty CompositionStringProperty =
            DependencyProperty.Register("CompositionString", typeof(string), typeof(Keyboard), new PropertyMetadata(string.Empty));

        public Keyboard()
        {
            InitializeComponent();

            var clickBinding = new CommandBinding(ClickCmd, OnExecuted, CanExecute);
            CommandManager.RegisterClassCommandBinding(typeof(Keyboard), clickBinding);

            this.InitializeKeyboard();
            this.Candidates = new ObservableCollection<string>();

            this.InitializeUILessMode();
        }

        public void Release()
        {
            _threadMgr = null;
            _uiElementMgr = null;

            if (_releaser != null)
            {
                _releaser.Dispose();
            }
        }

        #region ITfUIElementSink

        public HRESULT BeginUIElement([In] uint dwUIElementId, [In, MarshalAs(UnmanagedType.Bool), Out] ref bool pbShow)
        {
            pbShow = false;

            return new HRESULT();
        }

        public HRESULT UpdateUIElement([In] uint dwUIElementId)
        {
            this.OnUpdateUIElent(dwUIElementId);

            return new HRESULT();
        }

        public HRESULT EndUIElement([In] uint dwUIElementId)
        {
            this.Candidates.Clear();
            return new HRESULT();
        }

        private void OnUpdateUIElent(uint dwUIElementId)
        {
            ITfUIElement uiElement = null;

            try
            {
                _uiElementMgr.GetUIElement(dwUIElementId, out uiElement);
                var candidateListUIElement = uiElement as ITfCandidateListUIElement;

                this.Candidates.Clear();

                uint candidateCount;
                candidateListUIElement.GetCount(out candidateCount);

                uint pageCount;
                candidateListUIElement.GetPageIndex(null, 0, out pageCount);

                if (pageCount == 0)
                {
                    return;
                }

                var pageStarts = new uint[pageCount];
                candidateListUIElement.GetPageIndex(pageStarts, pageCount, out pageCount);

                uint currentPage;
                candidateListUIElement.GetCurrentPage(out currentPage);

                var end = candidateCount;
                var start = pageStarts[currentPage];

                if (currentPage != pageCount - 1)
                {
                    end = pageStarts[currentPage + 1];
                }

                for (var i = start; i < end; ++i)
                {
                    string candidate;
                    candidateListUIElement.GetString(i, out candidate);
                    this.Candidates.Add(candidate);
                }
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex);
            }
            finally
            {
                if (uiElement != null)
                {
                    Marshal.ReleaseComObject(uiElement);
                }
            }
        }

        #endregion

        #region ITfTextEditSink

        public HRESULT OnEndEdit([In, MarshalAs(UnmanagedType.Interface)] ITfContext pic, [In] uint ecReadOnly, [In, MarshalAs(UnmanagedType.Interface)] ITfEditRecord pEditRecord)
        {
            Debug.WriteLine("InputMethodProcessor.OnEndEdit");

            this.CheckComposition(pic, ecReadOnly);

            return new HRESULT();
        }

        private void CheckComposition(ITfContext pic, uint ecReadOnly)
        {
            this.CompositionString = string.Empty;

            var contextCompostion = pic as ITfContextComposition;
            if (contextCompostion == null) return;

            IEnumITfCompositionView enumCompositionView;
            var result = contextCompostion.EnumCompositions(out enumCompositionView);
            if (!result) return;

            var compositionView = new ITfCompositionView[1];

            try
            {
                var fetchCount = 0U;
                var resultStr = string.Empty;
                while (enumCompositionView.Next(1, compositionView, out fetchCount) && fetchCount > 0)
                {
                    ITfRange range;
                    result = compositionView[0].GetRange(out range);
                    if (result)
                    {
                        var text = new char[256];
                        var writeCount = 0U;
                        //bool isEmpty;
                        //range.IsEmpty(ecReadOnly, out isEmpty);
                        range.GetText(ecReadOnly, TF_TF.TF_TF_MOVESTART, text, 255, out writeCount);
                        var tempStr = new string(text).Trim('\0');
                        resultStr += tempStr;
                        Debug.WriteLine("resultStr={0}, tempStr={1}", resultStr, tempStr);
                    }

                    Marshal.ReleaseComObject(compositionView[0]);
                }

                textblock.Text = resultStr;
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex);
            }
            finally
            {
                Marshal.ReleaseComObject(enumCompositionView);
            }
        }

        #endregion

        #region ITfThreadMgrEventSink

        public HRESULT OnInitDocumentMgr([In, MarshalAs(UnmanagedType.Interface)] ITfDocumentMgr pdim)
        {
            return new HRESULT();
        }

        public HRESULT OnUninitDocumentMgr([In, MarshalAs(UnmanagedType.Interface)] ITfDocumentMgr pdim)
        {
            return new HRESULT();
        }

        public HRESULT OnSetFocus([In, MarshalAs(UnmanagedType.Interface)] ITfDocumentMgr pdimFocus, [In, MarshalAs(UnmanagedType.Interface)] ITfDocumentMgr pdimPrevFocus)
        {
            Debug.WriteLine("InputMethodProcessor.OnSetFocus");
            this.InitTextEditSink(pdimFocus);

            return new HRESULT();
        }

        public HRESULT OnPushContext([In, MarshalAs(UnmanagedType.Interface)] ITfContext pic)
        {
            return new HRESULT();
        }

        public HRESULT OnPopContext([In, MarshalAs(UnmanagedType.Interface)] ITfContext pic)
        {
            return new HRESULT();
        }

        private void InitTextEditSink(ITfDocumentMgr pdimPrevFocus)
        {
            if (_textEditSinkCookie != TextFrameworkDeclarations.TF_INVALID_COOKIE)
            {
                var source = _textEditSinkContext as ITfSource;
                source?.UnadviseSink(_textEditSinkCookie);

                Marshal.ReleaseComObject(_textEditSinkContext);
                _textEditSinkContext = null;
                _textEditSinkCookie = TextFrameworkDeclarations.TF_INVALID_COOKIE;
            }

            if (pdimPrevFocus == null)
                return;

            pdimPrevFocus.GetTop(out _textEditSinkContext);
            if (_textEditSinkContext != null)
            {
                var source = _textEditSinkContext as ITfSource;
                var guid = typeof(ITfTextEditSink).GUID;
                source?.AdviseSink(guid, this, out _textEditSinkCookie);
            }
        }

        #endregion

        private static void OnExecuted(object sender, ExecutedRoutedEventArgs e)
        {
            var key = (VirtualKeyCode)Enum.Parse(typeof(VirtualKeyCode), "VK_" + e.Parameter.ToString());
            InputSimulator.SimulateKeyPress(key);
        }

        private static void CanExecute(object sender, CanExecuteRoutedEventArgs e)
        {
            e.CanExecute = true;
        }

        private void InitializeUILessMode()
        {
            _releaser = new ComReleaser();
            _threadMgr = _releaser.CreateComObject<ITfThreadMgrEx>();
            _uiElementMgr = _threadMgr as ITfUIElementMgr;

            uint clientId;
            _threadMgr.ActivateEx(out clientId, TF_TMAE.TF_TMAE_UIELEMENTENABLEDONLY);

            var source = _uiElementMgr as ITfSource;
            var guid = typeof(ITfUIElementSink).GUID;
            uint uiElementSinkCookie;
            source.AdviseSink(ref guid, this, out uiElementSinkCookie);

            //var documentMgr = _releaser.ReceiveObject((out ITfDocumentMgr _) => _threadMgr.CreateDocumentMgr(out _));
            //ITfContext context;
            //documentMgr.CreateContext(clientId, 0, this, out context, out _editCookie);
            //documentMgr.Push(context);

            guid = typeof(ITfThreadMgrEventSink).GUID;
            source.AdviseSink(guid, this, out _threadMgrEventSinkCookie);

            //ITfDocumentMgr documentMgr;
            //_threadMgr.GetFocus(out documentMgr);
            //this.InitTextEditSink(documentMgr);
            //Marshal.ReleaseComObject(documentMgr);
        }

        private void InitializeKeyboard()
        {
            firstRowListBox.ItemsSource = new string[] { "Q", "W", "E", "R", "T", "Y", "U", "I", "O", "P" };
            secondRowListBox.ItemsSource = new string[] { "A", "S", "D", "F", "G", "H", "J", "K", "L" };
            thirdRowListBox.ItemsSource = new string[] { "Z", "X", "C", "V", "B", "N", "M" };
        }

        public HRESULT OnStartComposition([In, MarshalAs(UnmanagedType.Interface)] ITfCompositionView pComposition, [MarshalAs(UnmanagedType.Bool), Out] out bool pfOk)
        {
            pfOk = true;

            return new HRESULT();
        }

        public HRESULT OnUpdateComposition([In, MarshalAs(UnmanagedType.Interface)] ITfCompositionView pComposition, [In, MarshalAs(UnmanagedType.Interface)] ITfRange pRangeNew)
        {
            bool isEmpty;
            pRangeNew.IsEmpty(0, out isEmpty);

            return new HRESULT();
        }

        public HRESULT OnEndComposition([In, MarshalAs(UnmanagedType.Interface)] ITfCompositionView pComposition)
        {
            return new HRESULT();
        }
    }
}
